#include <windows.h>
#include "httpext.h"
#include <stdio.h>

#include "creg.hxx"

/// Useful macros
#define _stricmp(sz1, sz2) _memicmp(sz1, sz2, 1+min(strlen(sz1), strlen(sz2)))
#define strcmpi _stricmp

///  Registry keys
#define RK_HTTPD        L"COMM\\HTTPD"
#define RK_HTTPDVROOTS  L"COMM\\HTTPD\\VROOTS"


///  Normally we'd read this data from a file, allowing only the file 
///  to be changed.  To keep things simple, though, we just hard code it here.


///  This is the default page for the administration tool
const char cszWelcomePage[] =   "\
<html>  \r\n\
<title>Microsoft Windows CE Web Server</title>  \r\n\
</head> \r\n\
<body>  \r\n\
\r\n\
<p><B>Welcome to the Microsoft Windows CE Web Server!</B>   \r\n\
<BR><a href=\"/isapi/users\">Add/Remove Users.</a>  \r\n\
<BR><a href=\"/isapi/restart\">Stop and restart the web server.</a> \r\n\
<BR><a href=\"/isapi/regflush\">Flush Registry.</a> \r\n\
</body> \r\n\
</html> \r\n\
";

/// This is the top of the Users page
const char cszUsersTop[] =  "\
<html>  \r\n\
<title>Microsoft Windows CE Web Server</title>  \r\n\
</head> \r\n\
<body>  \r\n\
<B>View, Add, Modify and Delete Users</B> \r\n\
<TABLE BORDER=1 WIDTH=80%>  \r\n\
<THEAD> \r\n\
<TR>    \r\n\
<TH>User Name</TH>  \r\n\
<TH>Password</TH>   \r\n\
<TH>Modify</TH> \r\n\
<TH>Delete</TH> \r\n\
</TR>   \r\n\
<TBODY> \r\n\
";

/// This is the string output for each User row.
const char cszUserFormat[] = "\
<TR>    \r\n\
<FORM ACTION = \"/ISAPI/users\" METHOD = \"post\">  \r\n\
<TD><INPUT NAME = \"UserName\"  READONLY VALUE = \"%s\">  </TD> \r\n\
<TD><INPUT NAME = \"Password\" VALUE = \"xxxxxxxx\"></TD>   \r\n\
<TD><INPUT TYPE = submit NAME=\"Action\" VALUE=\"Modify\"></TD>  \r\n\
<TD><INPUT TYPE = submit NAME=\"Action\" VALUE=\"Delete\"></TD>  \r\n\
</FORM>  \r\n\
</TR>  \r\n\
";

///  This is displayed at the bottom of the Users page.  
const char cszUsersBottom[] = "\
<!-- The last row in this table is used to Add new Users. -->   \r\n\
<TR>  \r\n\
<FORM ACTION = \"/ISAPI/users\" METHOD = \"post\">  \r\n\
<TD><INPUT NAME = \"UserName\"  value = \"\"</TD>  \r\n\
<TD><INPUT NAME = \"Password\" value = \"\"</TD>  \r\n\
<TD><INPUT TYPE = submit NAME=\"Action\" VALUE=\"Add\"></TD>  \r\n\
</FORM>  \r\n\
</TR>  \r\n\
</TABLE>  \r\n\
<BR><a href=\"/ISAPI\">Go to main page.</a>  \r\n\
</body>  \r\n\
</html>  \r\n\
";  

///  Sent to the client when restarting the web server
const char cszRestartPageSuccess[] = "\
<html>  \r\n\
<title>Microsoft Windows CE Web Server</title>  \r\n\
</head> \r\n\
<body>  \r\n\
<B>Resetting the Windows CE Web Server.</B>   \r\n\
<P>The Web Server is shutting down.  It will not accept any more  \r\n\
connections until all open connections have been terminated. \r\n\
<BR><a href=\"/ISAPI\">Go to main page.</a>  \r\n\
</body>  \r\n\
</html>  \r\n\
";

const char cszRestartPageFailed[] = "\
<html>  \r\n\
<title>Microsoft Windows CE Web Server</title>  \r\n\
</head> \r\n\
<body>  \r\n\
<B>Resetting the Windows CE Web Server.</B>   \r\n\
<P><B>Error:</B> The Web Server cannot be shutdown at this time.  Reason unknown.\r\n\
<BR><a href=\"/ISAPI\">Go to main page.</a>  \r\n\
</body>  \r\n\
</html>  \r\n\
";

///  Sent to the client after flushing registry
const char cszRegFlushedPage[] = "\
<html>  \r\n\
<title>Microsoft Windows CE Web Server</title>  \r\n\
</head> \r\n\
<body>  \r\n\
<B>Registry has been Flushed</B>   \r\n\
<BR><a href=\"/ISAPI\">Go to main page.</a>  \r\n\
</body>  \r\n\
</html>  \r\n\
";

DWORD DisplayUsers(LPEXTENSION_CONTROL_BLOCK pECB);
DWORD RestartServer(LPEXTENSION_CONTROL_BLOCK pECB);
DWORD RegFlush(LPEXTENSION_CONTROL_BLOCK pECB);
DWORD DisplayWelcomePage(LPEXTENSION_CONTROL_BLOCK pECB);

void UpdateUsers(LPEXTENSION_CONTROL_BLOCK pECB);
char DecodeFromURL(char **pszSource, char *szDelimiters, char *szDest, UINT uCodePage);


/// What action to take on the virtual roots
typedef enum
{
    VROOT_UNKNOWN,
    VROOT_MODIFY,
    VROOT_ADD,
    VROOT_DELETE
} 
VROOT_ACTION;

extern "C" BOOL NTLMDeleteUser(IN LPTSTR pszUser);
extern "C" BOOL NTLMEnumUser(IN DWORD index, OUT LPTSTR pszUser, IN OUT LPDWORD pcchUser);
extern "C" BOOL NTLMSetUserInfo(IN LPTSTR pszUser, IN LPTSTR pszPassword);

extern "C" BOOL WINAPI GetExtensionVersion(HSE_VERSION_INFO* pVer)
{
    pVer->dwExtensionVersion = HSE_VERSION;
    return TRUE;
}

extern "C" DWORD WINAPI HttpExtensionProc(LPEXTENSION_CONTROL_BLOCK pECB)
{
    CHAR  szPathInfo[128];
    DWORD cbPathInfo = sizeof(szPathInfo);

    /// All requests to the admin tool go through this ISAPI dll.  To determine
    /// which page to return, we look at the PATH_INFO variable.

    /// In a request for http://devicename/ISAPI/users, the PATH_INFO will
    /// be /users.  This makes the URL easier for people to remember and
    /// type in than if we used a query string.

    /// Also note that the fact that this tool implemented as an ISAPI dll is 
    /// hidden from the user.  All requests that match /ISAPI are mapped
    /// to ISAPI.dll, rather than a directory, so there is to .dll anywhere
    /// in the path.

    if (! pECB->GetServerVariable(pECB->ConnID, "PATH_INFO", szPathInfo, &cbPathInfo))
        return DisplayWelcomePage(pECB);

    if (0 == strcmpi(szPathInfo, "/Users"))
        return DisplayUsers(pECB);
    else if (0 == strcmpi(szPathInfo, "/Restart"))
        return RestartServer(pECB);
    else if (0 == strcmpi(szPathInfo, "/RegFlush"))
        return RegFlush(pECB);
    else
        return DisplayWelcomePage(pECB);

    return HSE_STATUS_ERROR;   /// Defensive coding.  Should not reach this statement
}

DWORD DisplayWelcomePage(LPEXTENSION_CONTROL_BLOCK pECB)
{
    CHAR szHeaders[128];
    DWORD cbAdminPage = sizeof(cszWelcomePage) - 1;

    ///  This demonstrates how to use keep-alive in ISAPI extensions.
    ///  To do this, we must return HSE_STATUS_SUCCESS_AND_KEEP_CONN to the web
    ///  server.  When sending the http headers, we must set the Connection
    ///  to keep alive.  We also must include the Content-length, or the
    ///  browser wouldn't know when a request had ended.

    ///  Using keep-alive, if the browser supports them, will increase performance
    ///  of the web server because it will be able to use the same socket and internal
    ///  data structures for multiple request from the same client rather than 
    ///  constantly reinitializing them.

    sprintf(szHeaders,"Connection: keep-alive\r\n"
                      "Content-length: %d\r\n", cbAdminPage);
    pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, 0, 0, (LPDWORD) szHeaders);
    pECB->WriteClient(pECB->ConnID,(LPVOID) cszWelcomePage, &cbAdminPage,0);

    return HSE_STATUS_SUCCESS_AND_KEEP_CONN;
}
 
DWORD DisplayUsers(LPEXTENSION_CONTROL_BLOCK pECB)
{
    CHAR szBuf[2048];
    DWORD cbBuf;
    DWORD i;
    WCHAR wszUserName   [MAX_PATH+1];       /// virtual path 
    CHAR  szUserName    [MAX_PATH+1];       /// corresponding mapping to physical path
    DWORD cbUser = MAX_PATH;

    if (pECB->lpbData && pECB->lpbData[0] != 0)
        UpdateUsers(pECB);

    pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, 0, 0, 0);
    cbBuf = sizeof(cszUsersTop) - 1;
    pECB->WriteClient(pECB->ConnID, (LPVOID) cszUsersTop, &cbBuf, 0);

    for (i = 0; NTLMEnumUser(i, wszUserName, &cbUser); i++)
    { 
        WideCharToMultiByte(CP_ACP, 0, wszUserName,  -1, szUserName,  sizeof(szUserName),  0, 0);       

        /// Puts an entire row of the HTML table into szBuf.
        cbBuf = sprintf(szBuf, cszUserFormat, szUserName);
        pECB->WriteClient(pECB->ConnID, szBuf, &cbBuf, 0);
        cbUser = MAX_PATH;
    }

    cbBuf = sizeof(cszUsersBottom) - 1;
    pECB->WriteClient(pECB->ConnID,(LPVOID) cszUsersBottom, &cbBuf, 0); 
    
    return HSE_STATUS_SUCCESS;
}

void UpdateUsers(LPEXTENSION_CONTROL_BLOCK pECB)
{
    PSTR pszTrav = (PSTR) pECB->lpbData;        /// walks through the entry
    PSTR pszName;
    PSTR pszValue;
    WCHAR wszUserName[MAX_PATH+1];
    WCHAR wszPassword[MAX_PATH+1];
    

    VROOT_ACTION va = VROOT_UNKNOWN;

    /// The data will come to us in the form Name=Value&Name2=Value2
    /// We use DecodeFromURL because the data may have escape sequences in it,
    /// The CE Web Server gives us the data exactly as it was received from the client.
    
    while ( *pszTrav != '\0')
    {
        DecodeFromURL( &pszTrav, "=", pszName = pszTrav,  CP_ACP);
        DecodeFromURL( &pszTrav, "&", pszValue = pszTrav, CP_ACP);      

        /// See what name we have and take the appropriate action.
        if (0 == strcmpi(pszName, "UserName"))
            MultiByteToWideChar(CP_ACP, 0, pszValue, -1, wszUserName, sizeof(wszUserName));
        else if (0 == strcmpi(pszName, "Password"))
            MultiByteToWideChar(CP_ACP, 0, pszValue, -1, wszPassword, sizeof(wszPassword));
        else if (0 == strcmpi(pszName, "Action"))
        {
            if (0 == strcmpi(pszValue, "Modify"))
                va = VROOT_MODIFY;
            else if (0 == strcmpi(pszValue, "Add"))
                va = VROOT_ADD;
            else if (0 == strcmpi(pszValue, "Delete"))
                va = VROOT_DELETE;
        }
    }

    /// If we don't know what action to perform, cancel the operation
    if (VROOT_UNKNOWN == va)
        return;

    if (VROOT_DELETE == va)
    {
        NTLMDeleteUser(wszUserName);        
        return;
    }

    /// MODIFY or ADD is the same...
    NTLMSetUserInfo(wszUserName, wszPassword);

    return;
}


DWORD RestartServer(LPEXTENSION_CONTROL_BLOCK pECB)
{
    DWORD cbRestartPage;
    BOOL  fSuccess = FALSE;
    
    HINSTANCE hInst = LoadLibrary(L"httpd.dll");
    FARPROC ReInit;

    if (!hInst)
        goto done;

    ReInit = GetProcAddress(hInst,L"HttpdStopAndRestart");
    if ( ! ReInit)
        goto done;

    typedef void (WINAPI *PFN_EXECUTE)();
    ((PFN_EXECUTE) ReInit)();

    /// Note that even if we fail, we do NOT
    /// send a response code of "500 - Internal Error" in the 
    /// ServerSupportFunction below. Instead we return the default
    /// code (which is "200 - OK") and an informative message.
    
    fSuccess = TRUE;
done:
    cbRestartPage = fSuccess ? sizeof(cszRestartPageSuccess) - 1:       
                               sizeof(cszRestartPageFailed) - 1;
                               

    pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, NULL, NULL,0);
    pECB->WriteClient(pECB->ConnID,(LPVOID) 
                     (fSuccess ? cszRestartPageSuccess : cszRestartPageFailed),
                     &cbRestartPage,0);

    if (hInst)
        FreeLibrary(hInst);

    /// Do NOT return HSE_STATUS_SUCCESS_KEEP_CONN.  We want all connections to terminate
    /// as soon as possible on restarting, as the web server 
    return HSE_STATUS_SUCCESS;
}


DWORD RegFlush(LPEXTENSION_CONTROL_BLOCK pECB)
{
    DWORD cbRegFlushedPage;
        
    RegFlushKey(HKEY_LOCAL_MACHINE);
    cbRegFlushedPage = sizeof(cszRegFlushedPage) - 1;
                               

    pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, NULL, NULL,0);
    pECB->WriteClient(pECB->ConnID,(LPVOID)cszRegFlushedPage, &cbRegFlushedPage, 0);


    /// Do NOT return HSE_STATUS_SUCCESS_KEEP_CONN.  We want all connections to terminate
    /// as soon as possible on restarting, as the web server 
    return HSE_STATUS_SUCCESS;
}

///  Utility functions
/*===================================================================
HexToChar

Convert two digit hex string to a hex byte

Parameters:
    szHex - pointer to two digit hex string

Return Value:
    the character value of the hex string
===================================================================*/

char HexToChar(LPSTR szHex)
{
    char chResult, chDigit;

    chDigit = toupper(szHex[0]);
    chResult = (chDigit >= 'A'? (chDigit - 'A' + 0xA) : (chDigit - '0')) << 4;

    chDigit = toupper(szHex[1]);
    chResult |= chDigit >= 'A'? (chDigit - 'A' + 0xA) : (chDigit - '0');

    return chResult;
}


/*===================================================================
DecodeFromURL

Convert two digit hex string to a hex byte

WARNING:  This function modifies the passed pszSource!!

Parameters:
    pszSource    - in/out parameter points to a substring in the URL which
                   contains a Name=Value pair
    szDelimiters - a set of delimiters for this field
    szDest       - pointer to buffer to hold the substring

Return Value:
    Returns the actual delimiter that caused parsing to halt.
===================================================================*/

char DecodeFromURL(char **pszSource, char *szDelimiters, char *szDest, UINT uCodePage)
{
    char  ch;
    char *szSource = *pszSource;
    char *pszDestStart = szDest;
    CPINFO  CpInfo;
    BOOL   fIschLeadingByte = TRUE;

    GetCPInfo(uCodePage, (LPCPINFO)&CpInfo);

    while ((ch = *szSource++) != '\0' && ((!strchr(szDelimiters, ch) && fIschLeadingByte) || (!fIschLeadingByte)))
        switch (ch)
        {
            case ' ':       /// skip whitespace - assume that all whitespace
            case '\t':      /// that we need is escaped
            case '\r':      /// all these chars are out of trailing byte range
            case '\n':
            case '\f':
            case '\v':
                continue;
            case '+':       /// '+' is out of trailing byte range, can never be a trailing byte
                *szDest++ = ' ';
                break;
            case '%':       /// '%' is out of trailing byte range, can never be a trailing byte
                if (*szSource == 'u')
                {
                    WCHAR   wch;
                    wch =  (UCHAR)HexToChar(&szSource[1]) << 8;
                    wch |= (UCHAR)HexToChar(&szSource[3]);
                    szSource += 5;
                    szDest += WideCharToMultiByte( CP_ACP, 0, &wch, 1, szDest, 2,NULL, NULL );
                    break;
                }
                else
                {
                    ch = HexToChar(szSource);
                    szSource += 2;
                }
                /// FALL THROUGH to "Normal" case
            default:
                if (fIschLeadingByte == TRUE)
                {
                if (CpInfo.MaxCharSize > 1)
                    {
                        /// if this is a Leading byte, then, the next one is a trailing byte, we need
                        /// not process the next char even the next char is in szDelimiter, next char
                        /// is just the second byte of a DBCS char.
                        if (IsDBCSLeadByteEx(uCodePage, ch))
                            fIschLeadingByte = FALSE;
                    }
                }
                else
                {   /// A trailing byte
                    /// If we skip a DBCS trailing byte, then, the next char we check is a leading byte
                    fIschLeadingByte = TRUE;
                }
                *szDest++ = ch;
         }

    if (ch == '\0')     ///< End of String - undo increment of szSource
        --szSource;

    *szDest = '\0';

    *pszSource = szSource;
    return ch;
}
